# encoding: utf-8

"""
This module provides unified interface to anonymize data values.
"""
import logging
import time
import warnings

import numpy as np

from itertools import combinations, product
from multiprocessing import Pool, freeze_support

from pandas import DataFrame, merge
from numpy import random

from scipy.optimize import fsolve

from privacy.attribute import DataType
from privacy.dataset import DataSet
from privacy.utils import (set_random_seed, mutual_information,
                           normalize_distribution)

logger = logging.getLogger(__name__)


class DifferentialPrivacy(object):
    """
    The class to anonymize data set by a bayesian network to infer attributes'
    dependence relationship and differential privacy to keep differentially
    private.
    """
    def __init__(self, epsilon=0.1, degree=2, bins=20,
                 categories=None, pseudonyms=None, deletes=None, retains=None):
        """
        Constructor for DifferentialPrivacy class.

        Parameters
        ----------
        epsilon: float, default 0.1
            parameter for laplace mechanism

        categories : list of column names
            manually set categorical attributes

        pseudonyms : list of column names
            the columns to do pseudonym

        deletes : list of column names
            the columns to delete

        retains : list of column names
            the columns to retain

        degree : int, default 2
            degree of bayesian network

        bins : int, or string, or sequence
            used for numpy.histogram. only 'int' is used in this project.
        """
        self._epsilon = epsilon
        self._deletes = deletes or []
        self._categories = categories or []
        self._pseudonyms = pseudonyms or []
        self._retains = retains or []
        self._degree = degree
        self._bins = bins

        self._network = None
        self._bn_nodes = None

        self._dataset = None
        self._anonymized = None
        self._columns = None  # list of column names in data set

    def anonymize(self, data: DataFrame):
        """
        Anonymize a data frame.

        Parameters
        ----------
        data : pandas.DataFrame
            data set to be anonymized
        """
        self._dataset = DataSet(data, bins=self._bins,
                                categories=self._categories)
        # If delete-attributes is provided, anonymized csv don't include them.
        self._columns = [col for col in data.columns if col not in self._deletes]

        # nodes for bayesian networks
        self._bn_nodes = set()
        for col in self._columns:
            if col in self._pseudonyms:
                continue
            if (self._dataset[col].dtype == DataType.STRING and
                    not self._dataset[col].categorical):
                continue
            self._bn_nodes.add(col)
        # main steps of private bayesian network for anonymization
        # self._inject_laplace_noises(epsilon=self.epsilon)
        self._infer_bayesian_network(epsilon=self._epsilon, k=self._degree)
        self._correlated_anonymize()

        return self._anonymized

    def _infer_bayesian_network(self, k=0, epsilon=0.1):
        # encode dataset into bin indexes for bayesian network
        encoded_df = DataFrame()
        for col in self._bn_nodes:
            encoded_df[col] = self._dataset[col].to_indexes()

        if encoded_df.shape[1] < 2:
            raise Exception('If infer bayesian network, it requires at least 2 '
                            'attributes in dataset.')

        start = time.time()
        # bayesian network is defined as a set of AP (attribute-parent) pairs.
        # e.g. [(x1, p1), (x2, p2), ...], and pi is the parent set of xi
        self._network = greedy_bayes(encoded_df, k, epsilon=epsilon / 2,
                                     retains=self._retains)
        logger.info('greedy_bayes: %s', time.time() - start)
        self._cond_prs = noisy_conditionals(self._network,
                                            encoded_df, epsilon / 2)

    def _inject_laplace_noises(self, epsilon=0.1):
        size = len(self._bn_nodes)
        for col in self._columns:
            attr = self._dataset[col]
            attr.laplace(epsilon, size)

    def _independent_anonymize(self, seed=0):
        set_random_seed(seed)
        self._anonymized = DataFrame(columns=self._columns)
        for col in self._columns:
            attr = self._dataset[col]
            if not attr.categorical:
                self._anonymized[col] = attr.to_random()
            else:
                self._anonymized[col] = attr.choice()

    def _correlated_anonymize(self, seed=0):
        set_random_seed(seed)
        encoded_df = self._encode_dataset(self._dataset.n_rows)
        self._anonymized = DataFrame(columns=self._columns)
        for col in self._columns:
            attr = self._dataset[col]
            if col in self._pseudonyms:
                self._anonymized[col] = attr.to_pseudonym()
                continue
            if col in self._retains:
                self._anonymized[col] = attr.data
                continue
            if col in encoded_df:
                self._anonymized[col] = attr.choice(indexes=encoded_df[col])
            elif not self._dataset[col].categorical:
                # TODO regenerate pseudonymly or remote depends on parameters
                self._anonymized[col] = attr.to_random()
            else:
                # for attributes not in BN or candidate keys, use independent attribute mode.
                self._anonymized[col] = attr.choice()

    def _encode_dataset(self, n):
        root_col = self._network[0][1][0]
        root_prs = self._cond_prs[root_col]

        def sampling_order(bn):
            order = [bn[0][1][0]]
            for node, _ in bn:
                order.append(node)
            return order

        # encoded DataFrame
        edf = DataFrame(columns=sampling_order(self._network))
        edf[root_col] = random.choice(len(root_prs), size=n, p=root_prs)

        for child, parents in self._network:
            child_cond_prs = self._cond_prs[child]
            for parents_instance in child_cond_prs.keys():
                dist = child_cond_prs[parents_instance]
                parents_instance = list(eval(parents_instance))

                filter_cond = ''
                for parent, value in zip(parents, parents_instance):
                    filter_cond += f"(edf['{parent}']=={value})&"

                filter_cond = eval(filter_cond[:-1])
                size = edf[filter_cond].shape[0]
                if size:
                    edf.loc[filter_cond, child] = random.choice(len(dist),
                                                                size=size,
                                                                p=dist)

            child_prs = self._dataset[child].prs
            edf.loc[edf[child].isnull(),
                    child] = random.choice(len(child_prs),
                                           size=edf[child].isnull().sum(),
                                           p=child_prs)
        edf[edf.columns] = edf[edf.columns].astype(int)
        return edf

# -----------------------------------------------------------------------------
# Algorithms for PrivBayes


def calculate_sensitivity(n_rows, child, parents, col_to_binary):
    """Sensitivity function for Bayesian network construction.
    Lemma 4.1 Page 12

    Parameters
    ----------
    n_rows : int
        Number of tuples in sensitive dataset.

    child : str
        one column name

    parents : tuple
        parents of child, there may be multiple parents

    col_to_binary : dict
        mapping from column name to is_binary

    Return
    --------
    int
        Sensitivity value.
    """
    from math import log
    if col_to_binary[child] or (len(parents) == 1 and col_to_binary[parents[0]]):
        a = log(n_rows) / n_rows
        b = (1 / n_rows - 1) * log(1 - 1 / n_rows)
        return a + b
    else:
        a = (2 / n_rows) * log((n_rows + 1) / 2)
        b = (1 - 1 / n_rows) * log(1 + 2 / (n_rows - 1))
        return a + b


def calculate_delta(n_cols, sensitivity, epsilon):
    """
    Computing delta, which is a factor when applying differential privacy.

    Parameters
    ----------
    n_cols : int
        Number of attributes in dataset.
    sensitivity : float
        Sensitivity of removing one tuple.
    epsilon : float
        Parameter of differential privacy.
    """
    return (n_cols - 1) * sensitivity / epsilon


def usefulness_minus_target(k, n_cols, n_tuples, target_usefulness=5, epsilon=0.1):
    """Usefulness function.
    Lemma 4.8 Page 16

    Parameters
    ----------
    k : int
        Max number of degree in Bayesian networks construction
    n_cols : int
        Number of attributes in dataset.
    n_tuples : int
        Number of tuples in dataset.
    target_usefulness : int or float
    epsilon : float
        Parameter of differential privacy.
    """
    if k == n_cols:
        usefulness = target_usefulness
    else:
        usefulness = n_tuples * epsilon / ((n_cols - k) * (2 ** (k + 2)))
    return usefulness - target_usefulness


def calculate_k(num_attributes, num_tuples, target_usefulness=4, epsilon=0.1):
    """Calculate the maximum degree when constructing Bayesian networks."""
    default_k = 3
    initial_usefulness = usefulness_minus_target(default_k, num_attributes,
                                                 num_tuples, 0, epsilon)
    if initial_usefulness > target_usefulness:
        return default_k
    else:
        from math import ceil
        arguments = (num_attributes, num_tuples, target_usefulness, epsilon)
        warnings.filterwarnings("error")
        try:
            ans = fsolve(usefulness_minus_target,
                         np.array([int(num_attributes / 2)]),
                         args=arguments)[0]
            ans = ceil(ans)
        except RuntimeWarning:
            print("Warning: k is not properly computed!")
            ans = default_k
        if ans < 1 or ans > num_attributes:
            ans = default_k
        return ans


def worker(paras):
    """
    Process task to calculate mutual information of attribute-parents pairs.
    """
    child, V, n_parents, split, dataset = paras
    aps = []
    mis = []

    if split + n_parents - 1 < len(V):
        for other_parents in combinations(V[split + 1:], n_parents - 1):
            parents = list(other_parents)
            parents.append(V[split])
            aps.append((child, parents))
            mi = mutual_information(dataset[child], dataset[parents])
            mis.append(mi)

    return aps, mis


def greedy_bayes(dataset: DataFrame, k=None, epsilon=None, retains=None):
    """
    Construct a Bayesian Network using greedy algorithm.
    Algorithm 4, Page 20

    Parameters
    ----------
    dataset : DataFrame
        Input dataset, which only contains categorical attributes.

    k : int
        Maximum degree of the constructed bayesian network. If k=0, then k is
        automatically calculated.

    epsilon : float
        Parameter of differential privacy.

    retains : list of columns
        the columns to retain
    """
    freeze_support()

    dataset: DataFrame = dataset.astype(str, copy=False)
    n_rows, n_cols = dataset.shape
    retains = retains or []
    if not k:
        k = calculate_k(n_cols, n_rows)

    # mapping from column name to is_binary, because sensitivity is different
    # for binary or non-binary column
    col_to_binary = {col: dataset[col].unique().size <= 2 for col in dataset}
    more_retains = False
    if len(retains) == 0:
        root_col = random.choice(dataset.columns)
    elif len(retains) == 1:
        root_col = retains[0]
    else:
        root_col = random.choice(retains)
        more_retains = True

    # V is a set which contains all attributes whose parent sets has been set
    V = [root_col]
    if more_retains:
        rest_cols = set(retains)
    else:
        rest_cols = set(dataset.columns)
    rest_cols.remove(root_col)
    # N is the bayesian network
    N = []
    while rest_cols:
        # ap: attribute-parent (AP) pair is a tuple. It is a node in bayesian
        # network, e.g. ('education', ['relationship']), there may be multiple
        # parents, depends on k (degree of bayesian network).
        aps = []
        # mi: mutual information (MI) of two features
        mis = []

        n_parents = min(len(V), k)
        # calculate
        tasks = [(child, V, n_parents, split, dataset) for child, split in
                 product(rest_cols, range(len(V) - n_parents + 1))]
        with Pool() as pool:
            res_list = pool.map(worker, tasks)

        for res in res_list:
            aps += res[0]
            mis += res[1]

        # find next child node in bayesian networks according to the biggest
        # mutual information or exponential mechanism
        if epsilon:
            sample_prs = exponential_mechanism(epsilon, mis, aps,
                                               col_to_binary, n_rows, n_cols)
            index = np.random.choice(list(range(len(mis))), p=sample_prs)
        else:
            index = mis.index(max(mis))

        N.append(aps[index])
        next_attr = aps[index][0]
        V.append(next_attr)
        rest_cols.remove(next_attr)
        if len(rest_cols) == 0 and more_retains:
            rest_cols = set(dataset.columns) - set(retains)
            more_retains = False
    return N


def exponential_mechanism(epsilon, mis, aps, col_to_binary, n_rows, n_cols):
    """
    Apply Exponential Mechanism to mutual information.

    Parameters
    ----------
    epsilon : float
        parameter in differential privacy

    mis : list
        list of mutual information

    aps : list
        list of attribute-parents pairs

    col_to_binary : dict
        mapping from column name to is_binary

    n_rows : int
        number of tuples of dataset

    n_cols : int
        number of attributes (columns) in dataset
    """
    deltas = []
    for (child, parents) in aps:
        sensitivity = calculate_sensitivity(n_rows, child, parents, col_to_binary)
        delta = calculate_delta(n_cols, sensitivity, epsilon)
        deltas.append(delta)

    mi_array = np.array(mis) / (2 * np.array(deltas))
    mi_array = np.exp(mi_array)
    mi_array = normalize_distribution(mi_array)
    return mi_array


def laplace_noise_scale(k, num_attributes, num_tuples, epsilon):
    """
    Laplace scale for noisy conditional distribution.
    """
    return 2 * (num_attributes - k) / (num_tuples * epsilon)


def noisy_distributions(attributes, encoded_dataset, epsilon=0.1):
    data = encoded_dataset.copy().loc[:, attributes]
    data['count'] = 1
    stats = data.groupby(attributes).sum()

    iterables = [range(int(encoded_dataset[attr].max()) + 1) for attr in attributes]
    full_space = DataFrame(columns=attributes, data=list(product(*iterables)))
    stats.reset_index(inplace=True)
    stats = merge(full_space, stats, how='left')
    stats.fillna(0, inplace=True)

    if epsilon:
        k = len(attributes) - 1
        n_tuples, n_columns = encoded_dataset.shape
        scale = laplace_noise_scale(k, n_columns, n_tuples, epsilon)
        laplace_noises = np.random.laplace(0, scale=scale, size=stats.index.size)
        stats['count'] += laplace_noises
        stats.loc[stats['count'] < 0, 'count'] = 0

    return stats


def noisy_conditionals(bayesian_network, encoded_dataset, epsilon=0.1):
    """
    Generation of noisy conditional distribution probability
    Algorithm 3, Page 20
    TODO: how about non-binary attributes?
    """
    # TODO confirm: below statement is right? k should be degree of bayesian network
    k = len(bayesian_network[-1][1])
    cond_prs = {}

    # first k+1 attributes
    root = bayesian_network[0][1][0]
    k1_attrs = [root]
    for child, _ in bayesian_network[:k]:
        k1_attrs.append(child)

    k1_noisy_prs = noisy_distributions(k1_attrs, encoded_dataset, epsilon)

    # generate noisy distribution of root attribute.
    root_stats = k1_noisy_prs.loc[:, [root, 'count']].groupby(root).sum()['count']
    cond_prs[root] = normalize_distribution(root_stats).tolist()

    for idx, (child, parents) in enumerate(bayesian_network):
        cond_prs[child] = {}

        if idx < k:
            stats = k1_noisy_prs.copy().loc[:, parents + [child, 'count']]
        else:
            stats = noisy_distributions(parents + [child], encoded_dataset, epsilon)

        stats = DataFrame(stats.loc[:, parents + [child, 'count']]
                          .groupby(parents + [child]).sum())
        if len(parents) == 1:
            for parent_instance in stats.index.levels[0]:
                dist = normalize_distribution(stats.loc[parent_instance]['count']).tolist()
                cond_prs[child][str([parent_instance])] = dist
        else:
            for parents_instance in product(*stats.index.levels[:-1]):
                dist = normalize_distribution(stats.loc[parents_instance]['count']).tolist()
                cond_prs[child][str(list(parents_instance))] = dist

    return cond_prs
