"""
Attributes and Data Types in data anonymization.
"""

import numpy as np

from enum import Enum
from bisect import bisect_right
from random import uniform
from typing import List, Union
from pandas import Series, concat, DataFrame
from dateutil.parser import parse
from datetime import datetime, timedelta

from privacy import utils


class DataType(Enum):
    INTEGER = 'Integer'
    FLOAT = 'Float'
    STRING = 'String'
    DATETIME = 'DateTime'

    def __eq__(self, other):
        """
        Define __eq__ method to compare a string and a DateType.
        e.g. 'String' == DateType.STRING
        """
        if isinstance(other, str):
            return self.value == other
        else:
            return self is other

    def __str__(self):
        return self.value


class Attribute(object):

    def __init__(self, data: Series, name=None, dtype=None,
                 categorical=False, bins=20):
        """
        A Series with extra information, e.g. categorical.

        Parameters
        ----------
        data : pandas.Series
            a Series

        name : str
            name of attribute. (default Series.name)

        dtype : str
            dtype of attribute. Default is inferring by apis from pandas. It
            should come from class DataType.

        categorical : bool
            set categorical label for attribute. If categorical, this attribute
            takes on a limited and fixed number of possible values. Examples:
            blood type, gender.

        bins : int
            number of intervals for distributions of this attribute
        """
        self.name = name or data.name
        # bins can be int (size of histogram bins), str (as algorithm name),
        self._bins: Union[int, str] = bins
        self.data: Series = data
        self.data_dropna: Series = self.data.dropna()
        self._hasna = self.data.size != self.data_dropna.size

        self.numerical = None
        self.dtype = None
        self.min = None
        self.max = None
        self._step = None

        # probability distribution (pr)
        self.bins: np.ndarray = None
        self.prs: np.ndarray = None

        # infer data type of data series
        if dtype is None:
            self._infer_data_type(data)
        else:
            # TODO
            pass

        if self.dtype == DataType.DATETIME:
            self._epoch = datetime(1970, 1, 1)
            self._to_seconds = lambda x: int((parse(x) - self._epoch)
                                             .total_seconds())

        is_unique = self.data.is_unique
        self.categorical = categorical or (
                self.dtype == DataType.STRING and not is_unique)
        self._set_domain()
        self._set_distribution()
        if self.dtype == DataType.INTEGER or self.dtype == DataType.FLOAT:
            self.numerical = True

    def _infer_data_type(self, data):
        """
        Infer this attribute's data type by pandas.api.types.infer_dtype
        """
        from pandas.api.types import infer_dtype
        from privacy.utils import is_datetime
        dtype = infer_dtype(data, skipna=True)
        if dtype == 'integer':
            dtype = DataType.INTEGER
        elif dtype == 'floating' or dtype == 'mixed-integer-float':
            dtype = DataType.FLOAT
        elif dtype == 'string':
            samples = data.dropna().sample(frac=0.2, replace=False)
            if all(samples.map(is_datetime)):
                dtype = DataType.DATETIME
            else:
                dtype = DataType.STRING
        self.dtype = dtype

    def _set_domain(self):
        """
        Compute domain (min, max, distribution bins) from input data
        """
        if self.dtype == DataType.STRING:
            self.data_dropna_len = self.data_dropna.astype(str).map(len)
            self.min = int(self.data_dropna_len.min())
            self.max = int(self.data_dropna_len.max())
            if self.categorical:
                self.bins = self.data_dropna.unique()
            else:
                self.bins = np.array([self.min, self.max])
        elif self.dtype == DataType.DATETIME:
            self.data_dropna = self.data_dropna.map(self._to_seconds)
            self.min = float(self.data_dropna.min())
            self.max = float(self.data_dropna.max())
            if self.categorical:
                self.bins = self.data_dropna.unique()
            else:
                self.bins = np.array([self.min, self.max])
                self._step = (self.max - self.min) / self._bins
        else:
            self.min = float(self.data_dropna.min())
            self.max = float(self.data_dropna.max())
            if self.categorical:
                self.bins = self.data_dropna.unique()
            else:
                self.bins = np.array([self.min, self.max])
                self._step = (self.max - self.min) / self._bins

    def domain(self):
        if self.categorical:
            return self.bins
        else:
            return [self.min, self.max]

    def counts(self, bins=None):
        if bins is None:
            return self._counts
        if self.categorical:
            distribution = self.data_dropna.value_counts()
            for value in set(bins) - set(distribution.index):
                distribution[value] = 0
            return np.array([distribution.get(b) for b in bins])
        else:
            if len(bins) == 1:
                return np.array([self.data_dropna.size])
            distribution = np.histogram(self.data_dropna, bins=bins)
            return distribution[0]

    def reset_domain(self, domain: List):
        """
        Set attribute's domain, includes min, max, frequency, and distribution.

        Parameters
        ----------
        domain : list
            domain of one attribute. For numerical or datetime attributes, it
            should be a list of two elements [min, max]; For categorical
            attributes, it should a list of potential values of this attribute.
        """
        # if a attribute is numerical and categorical and domain's length is
        # bigger than 2, take it as categorical. e.g. zip code.
        if self.numerical and self.categorical and len(domain) > 2 or self.categorical:
            self.min = min(domain)
            self.max = max(domain)
            self.bins = np.array(domain)
        elif self.numerical:
            self.min, self.max = domain
            self._step = (self.max - self.min) / self._bins
            self.bins = np.array([self.min, self.max])
        elif self.dtype == DataType.STRING:
            lengths = [len(i) for i in domain]
            self.min = min(lengths)
            self.max = max(lengths)
            self.bins = np.array(domain)
        self._set_distribution()

    def _set_distribution(self):
        if self.categorical:
            distribution = self.data_dropna.value_counts()
            for value in set(self.bins) - set(distribution.index):
                distribution[value] = 0
            distribution.sort_index(inplace=True)
            self._counts = distribution.values
            self.prs = utils.normalize_distribution(distribution)
            self.bins = np.array(distribution.index)
        else:
            # Note: hist, bin_edges = numpy.histogram(), all but the last bin
            # is half-open. If bins is 20, then len(hist)=20, len(bin_edges)=21
            if self.dtype == DataType.STRING:
                distribution = np.histogram(self.data_dropna_len,
                                            bins=self._bins)
                self.bins = distribution[1][:-1]
                self._counts = distribution[0]
                self.prs = utils.normalize_distribution(distribution[0])
            else:
                distribution = np.histogram(self.data_dropna,
                                            bins=self._bins,
                                            range=(self.min, self.max))
                self.bins = distribution[1][:-1]  # Remove the last bin edge
                self._counts = distribution[0]
                self.prs = utils.normalize_distribution(distribution[0])
                if self.dtype == DataType.INTEGER:
                    self.min = int(self.min)
                    self.max = int(self.max)

    def laplace(self, epsilon, size):
        """
        Inject laplace noises to this attribute's probability distribution.
        """
        if epsilon > 0:
            sensitivity = 2 / self.data.size
            privacy_budget = epsilon / size
            scale = sensitivity / privacy_budget
            noises = np.random.laplace(0, scale=scale, size=len(self.prs))
            noisy_distribution = self.prs + noises
            self.prs = utils.normalize_distribution(noisy_distribution)

    def to_indexes(self):
        """
        Encode values into bin indexes for Bayesian Network.
        """
        if self.categorical:
            v2idx = {value: idx for idx, value in enumerate(self.bins)}
            indexes = self.data.map(lambda x: v2idx[x], na_action='ignore')
        else:
            if self.dtype == DataType.DATETIME:
                indexes = self.data.map(
                    lambda x: bisect_right(self.bins, self._to_seconds(x)) - 1,
                    na_action='ignore')
            else:
                indexes = self.data.map(lambda x: bisect_right(self.bins, x) - 1,
                                        na_action='ignore')

        indexes.fillna(len(self.bins), inplace=True)
        return indexes.astype(int, copy=False)

    def to_json(self):
        """
        Encode attribution information in JSON format / Python dictionary.
        """
        return {
            "name": self.name,
            "dtype": self.dtype.value,
            "categorical": self.categorical,
            "min": self.min,
            "max": self.max,
            "bins": self.bins.tolist(),
            "prs": self.prs.tolist()
        }

    def to_pseudonym(self):
        """
        Return pseudonymized values for this attribute, which is used to
        substitute identifiable data with a reversible, consistent value.
        """
        if self.categorical:
            b2pseu = {b: utils.pseudonymise_string(b) for b in self.bins}
            return self.data.map(lambda x: b2pseu[x])

        if self.dtype == DataType.STRING or self.dtype == DataType.DATETIME:
            return self.data.map(utils.pseudonymise_string)
        elif self.dtype == DataType.INTEGER or self.dtype == DataType.FLOAT:
            return self.data.map(str).map(utils.pseudonymise_string)

    def to_random(self, n=None):
        """
        Return an random array with same length (usually used for
        non-categorical attribute).
        """
        if n is None:
            n = len(self.data)
        if self.dtype == DataType.FLOAT:
            rands = np.arange(self.min, self.max, (self.max - self.min) / n)
            np.random.shuffle(rands)
            return rands

        rands = np.arange(n)
        np.random.shuffle(rands)
        if self.dtype == DataType.INTEGER:
            return rands
        elif self.dtype == DataType.DATETIME:
            pass
        elif self.dtype == DataType.STRING:
            if self.min == self.max:
                length = self.min
            else:
                length = np.random.randint(self.min, self.max)
            vectorized = np.vectorize(lambda x:
                                      utils.generate_random_string(length))
            return vectorized(rands)

    def _uniform_sample(self, index: int):
        num_bins = len(self.bins)
        if index == num_bins:
            return np.nan
        elif self.categorical:
            return self.bins[index]
        elif index < num_bins - 1:
            return uniform(self.bins[index], self.bins[index + 1])
        else:
            # sample from the last interval where the right edge is missing in self.bins
            neg_2, neg_1 = self.bins[-2:]
            return uniform(neg_1, self.max)

    def choice(self, n=None, indexes=None):
        """
        Return a random sample based on this attribute's probability.
        If indexes and n are both set, ignore n.

        Parameters
        ----------
        n : int
            size of random sample

        indexes : array-like
            array of indexes in bins
        """
        if indexes is None:
            if n is None:
                n = len(self.data)
            indexes = Series(np.random.choice(len(self.prs), size=n, p=self.prs))
        column = indexes.apply(lambda x: self._uniform_sample(x))
        if self.dtype == DataType.DATETIME:
            # TODO need to keep the format as input
            def format_datetime(date):
                return '%d/%d/%d' % (date.month, date.day, date.year)
            column = column.apply(
                lambda x: format_datetime(self._epoch + timedelta(seconds=x)))
        elif self.dtype == DataType.INTEGER:
            column = column.round()
            if self._hasna:
                column[~column.isnull()] = column[~column.isnull()].astype(int)
            else:
                column = column.astype(int)
        elif self.dtype == DataType.STRING:
            if not self.categorical:
                column[~column.isnull()] = column[~column.isnull()].apply(
                    lambda x: utils.generate_random_string(int(x)))
        return column

    def transform(self, y=None):
        """
        Transform labels to normalized encoding.

        Parameters
        ----------
        y : array-like
            target values
        """
        if y is None:
            y = self.data
        if self.categorical:
            tr = DataFrame()
            for c in self.bins:
                tr[c] = y.apply(lambda y1: 1 if y1 == c else 0)
            return tr
        else:
            if self.dtype != DataType.STRING:
                if self.dtype == DataType.DATETIME:
                    y = self.data_dropna
                return y.apply(lambda y1:
                               (int((y1-self.min)/(self._step + 1e-8))+0.5)/self._bins)

    def inverse_transform(self, y=None):
        if self.categorical:
            idx = np.argmax(np.asarray(y), axis=0)
            return Series(list(map(lambda i: self.bins[i], idx)), name=self.name)
        else:
            return y.apply(lambda y1: int((y1*self._bins-0.5) * (self._step + 1e-8)) + self.min)
