"""
evaluate.py

"""

import argparse
import os
import time

from privacy.evaluator import BiFrame, EvaluationReport
from privacy.utils import read_data_from_csv, CustomFormatter,\
    plot_histogram, plot_heatmap, plot_confusion_matrix, str_to_list


def main():
    parser = argparse.ArgumentParser(
        description='Evaluate anonymized data set and compare with source data set.',
        formatter_class=CustomFormatter,
        add_help=False)
    # positional arguments
    parser.add_argument('source',
                        help='set file path of source (raw) dataset to be compare '
                             'with anonymized dataset, support csv files')
    parser.add_argument('target',
                        help='set file path of target (anonymized) dataset to '
                             'evaluate')

    # optional arguments
    group = parser.add_argument_group('general arguments')
    group.add_argument("-h", "--help", action="help",
                       help="show this help message and exit")
    group.add_argument('--na-values', metavar='LIST',
                       help='set additional values to recognize as NA/NaN; ('
                            'default null values are from pandas.read_csv)')
    group.add_argument('-o', '--output', metavar='FILE', default='report.html',
                       help='set output path for evaluation report; (default is '
                            '"report.html" under current work directory)')

    group = parser.add_argument_group('advanced arguments')
    group.add_argument('-t', '--test',
                       help='set test dataset for classification or regression '
                            'task; (default take 20 percent from source dataset)')
    group.add_argument('--class-label', metavar='LIST',
                       help='set column name as class label for classification '
                            'or regression task; supports one or multiple '
                            'columns (separated by comma)')

    args = parser.parse_args()
    start = time.time()

    na_values = str_to_list(args.na_values)
    class_labels = str_to_list(args.class_label)

    # check kinds of parameters
    args.output = os.path.join(os.getcwd(), args.output)
    # if output folder not exists, then create it.
    if not os.path.exists(os.path.dirname(args.output)):
        os.makedirs(os.path.dirname(args.output))

    def complement(attrs, full):
        return set(attrs or []) - set(full)

    # Initialization task:
    source = read_data_from_csv(args.source, na_values=na_values, header='infer')
    target = read_data_from_csv(args.target, na_values=na_values, header='infer')
    test = read_data_from_csv(args.test) if args.test is not None else None

    comp = complement(class_labels, source.columns)
    if comp:
        parser.exit(message=f'--class-label(s): {comp} are not in source file.')
    comp = complement(class_labels, target.columns)
    if comp:
        parser.exit(message=f'--class-label(s): {comp} are not in target file.')

    frame = BiFrame(source, target)
    report = EvaluationReport(title='Data Utility Evaluation Report')
    # to_dict, return an json like {index: [...], columns: [...], data: [[...]]}
    report.add_part('basic', frame.describe().to_dict('split'))

    # check data's distribution
    for col in frame.columns:
        bins, counts = frame.dist(col)
        svg = plot_histogram(bins, counts)
        report.add_part('dist', {'name': col, 'columns': bins, 'data': counts,
                                 'path': svg})

    # check data's correlation relationship
    source_mi, target_mi = frame.corr()
    source_svg = plot_heatmap(source_mi)
    target_svg = plot_heatmap(target_mi)
    report.add_part('corr', {'matrix': source_mi.to_dict('split'),
                             'path': source_svg})
    report.add_part('corr', {'matrix': target_mi.to_dict('split'),
                             'path': target_svg})

    # if user enable the classification task,
    if class_labels is not None:
        # TODO: run svm classifier, change to a separate plugin
        for col in class_labels:
            in_test = (test is not None and col in test) or (test is None)
            if in_test:
                # When class label in svm classify test data, try to match two
                # predicted result with the actual data, and so, there will be
                # two confusion matrix diagrams.
                try:
                    source_cm, target_cm = frame.classify(col, test=test)
                    vrange = (min(source_cm.values.min(), target_cm.values.min()),
                              max(source_cm.values.max(), target_cm.values.max()))
                    path = (
                        plot_confusion_matrix(source_cm, vrange=vrange,
                                              xlabel='raw', ylabel='actual'),
                        plot_confusion_matrix(target_cm, vrange=vrange,
                                              xlabel='anon', ylabel='actual'))
                    report.add_part('svm', {'column': col, 'path': path})
                except ValueError as e:
                    print(e)
            else:
                # If not, will compare two predicted result.
                try:
                    cm = frame.classify(col, test=test)
                    # make path's type: 1-tuple
                    path = (plot_confusion_matrix(cm, xlabel='anon',
                                                  ylabel='raw'),)
                    report.add_part('svm', {'column': col, 'path': path})
                except ValueError as e:
                    print(e)

    report_path = args.output
    report.to_html(report_path)

    duration = time.time() - start
    print(f'Evaluation report: {report_path} in {round(duration, 2)} seconds.')


if __name__ == '__main__':
    # For Testing
    main()
