"""
DataSet

"""
from numpy import random

import pandas as pd

from privacy.attribute import DataType, Attribute
from privacy.utils import generate_random_string


class DataSet(object):
    def __init__(self, frame: pd.DataFrame, types: list = None, bins=20,
                 categories=None):
        """
        Constructor of DataSet class

        Parameters
        ----------
        frame : pd.DataFrame
            data set

        types : list
            list of data types for each column. If not set, infer by values.

        bins : int
            counts of consecutive, non-overlapped intervals to represent column.

        categories : list
            categorical columns
        """
        self.bins = bins

        self.separator = '_'

        self._attrs = {}     # map column names to attributes
        self._encoders = {}  # map column names to encoders (for ml algorithms)

        # data metadata
        self.n_rows = frame.shape[0]
        self.n_cols = frame.shape[1]
        self._data = frame

        # If types is not provided, infer them;
        # dtypes = {}  # data types (mapping from column name to data type)
        if types is not None and len(types) > 0:
            if len(frame.columns) == len(types):
                self.dtypes = dict(zip(frame.colums, types))
                # TODO how to check data types quickly
            else:
                raise ValueError('The types doesnt match the frame.')
        # infer candidate and category attributes
        # TODO: how about candidate attributes from input
        categories = categories or []
        for col in frame.columns:
            categorical = col in categories
            self._attrs[col] = Attribute(frame[col], categorical=categorical)

    def __getitem__(self, column):
        """
        Equals to self[column]
        """
        return self._attrs[column]

    def reset_domains(self, domains):
        """
        Reset every column's domain and distributions. If attribute's domain is
        reset, its distribution will be changed.
        Usually used to analyze two dataset based on same domains.
        """
        for col in domains:
            # set input parameter to infer domain, e.g. categorical domain, numerical range
            self._attrs[col].reset_domain(domains[col])

    def domain(self, column):
        return self._attrs[column].domain()

    def counts(self, column, bins=None):
        return self._attrs[column].counts(bins=bins)

    def dtype(self, column):
        return self._attrs[column].dtype

    def categorical(self, column):
        """
        Detect categorical column.
        Return a boolean to indicate whether the values of 'column' is categorical.
        """
        return self._attrs[column].categorical

    def mi(self):
        """ Return mutual information of all attributes. """
        from privacy.metrics import pairwise_mutual_information
        return pairwise_mutual_information(self._data)

    def random_like(self, seed=0):
        frame = pd.DataFrame()
        for col in self._attrs:
            dtype = self._attrs[col].dtype
            attr = self._attrs[col]
            if attr.categorical:
                frame[col] = random.choice(attr.bins, self.n_rows)
            elif dtype == DataType.STRING:
                length = random.randint(attr.min, attr.max)
                frame[col] = ([length] * length).map(lambda x: generate_random_string(x))
            else:
                if dtype == DataType.INTEGER:
                    frame[col] = random.randint(attr.min, attr.max + 1, self.n_rows)
                else:
                    frame[col] = random.uniform(attr.min, attr.max, self.n_rows)
        return frame

    def transform(self, data=None):
        """
        Transform data set to values by kinds of encoders.
        If data is set, use this data set's encoders to transform.
        """
        # If the data to transform is None, then transform source data _data;
        encoded = pd.DataFrame()
        for col in self._attrs:
            # TODO: Only focus on Bayesian Network nodes ??
            if data is not None and col not in data:
                continue
            attr = self._attrs[col]
            if attr.dtype == DataType.STRING and attr.categorical:
                tr_ = attr.transform(None if data is None else data[col])
                for label in attr.bins:
                    encoded[col + self.separator + label] = tr_[label]
            elif attr.dtype != DataType.STRING:
                encoded[col] = attr.transform(None if data is None else data[col])
        return encoded

    def inverse_transform(self, data):
        df = pd.DataFrame()
        for col in self._attrs:
            attr = self._attrs[col]
            if attr.categorical:
                values = list(data[col + self.separator + c] for c in attr.bins)
                df[col] = attr.inverse_transform(values)
            else:
                df[col] = attr.inverse_transform(data[col])
        return df
