"""
BiFrame for data anonymization.
"""

import logging
import numpy as np
import pandas as pd

from pandas import Index
from sklearn.model_selection import train_test_split

from privacy.dataset import DataSet
from privacy.utils import predict, normalize_range
from privacy.metrics import jensen_shannon_divergence, relative_error

logger = logging.getLogger(__name__)


class BiFrame(object):
    def __init__(self, left: pd.DataFrame, right: pd.DataFrame, metadata=None):
        """
        BiFrame class that contains two data sets: left and right DataFrame.
        It currently provides kinds of analysis methods from distribution,
        correlation, and some machine learning tasks.
        Especially, if the input left and right data set are source and
        anonymized dataset, this class can be used to evaluate the utility and
        privacy of anonymized data set.

        Parameters
        ----------
        left : {pandas.DataFrame}
            one data set

        right : {pandas.DataFrame}
            another data set, i.e. anonymized ata

        metadata : dict
            metadata information of data
        """
        # distribution
        self._dt = {}

        if metadata is not None and 'types' in metadata:
            types = metadata['types']
        else:
            types = []
        # To compare two data set, make sure that they have same columns.
        # If not, compare the common part.
        common = set(left.columns) & set(right.columns)
        if len(common) != len(left.columns) or len(common) != len(right.columns):
            logger.info(f"BiFrame constructed on attributes: {common}.")

        # left and right data frame (df)
        self._lt_df = left[common]
        self._rt_df = right[common]

        # left and right data set (ds)
        self._lt_ds = DataSet(self._lt_df, types=types)
        self._rt_ds = DataSet(self._rt_df, types=types)
        self._columns = self._lt_df.columns.sort_values().to_list()

        # Make sure that two dataset have same domain for categorical
        # attributes, and same min, max values for numerical attributes.
        domains = {}
        for col in self._columns:
            # If current column is not categorical, will ignore it.
            if not self._lt_ds[col].categorical or not self._rt_ds[col].categorical:
                continue
            d1, d2 = self._lt_ds.domain(col), self._rt_ds.domain(col)
            if not np.array_equal(d1, d2):
                if self._lt_ds.categorical(col):
                    domains[col] = np.unique(np.concatenate((d1, d2)))
                else:
                    domains[col] = [min(d1[0], d2[0]), max(d1[1], d2[1])]
        if len(domains) > 0:
            self._lt_ds.reset_domains(domains)
            self._rt_ds.reset_domains(domains)

    @property
    def columns(self):
        return self._columns

    def err(self):
        """
        Compute pairwise err (relative error) of columns' distribution.
        """
        # merge two frequency counts, and calculate relative difference
        df = pd.DataFrame(columns=self._columns, index=['err'])
        df.fillna(0)
        for col in self._columns:
            df.at['err', col] = relative_error(self._lt_ds[col].counts(),
                                               self._rt_ds[col].counts())
        return df

    def jsd(self):
        """
        Compute pairwise JSD (Jensen-Shannon divergence) of columns' distribution.
        """
        df = pd.DataFrame(columns=self._columns, index=['jsd'])
        df.fillna(0)
        for col in self._columns:
            df.at['jsd', col] = jensen_shannon_divergence(self._lt_ds[col].counts(),
                                                          self._rt_ds[col].counts())
        return df

    def corr(self):
        """
        Computer pairwise correlation and dependence measured by mi (mutual
        information).
        """
        return self._lt_ds.mi(), self._rt_ds.mi()

    def dist(self, column):
        """
        Return frequency distribution of column.

        Parameters
        ----------
        column : str
            column name, whose distribution will be return
        """
        if len(self._dt) == 0:
            for c in self._columns:
                self._dt[c] = {}
                if self._lt_ds.categorical(c):
                    bins = self._lt_ds.domain(c)
                    counts1 = self._lt_ds.counts(c, bins=bins)
                    counts2 = self._rt_ds.counts(c, bins=bins)
                else:
                    min_, max_ = self._lt_ds.domain(c)
                    # the domain from two data set are same;
                    # extend the domain to human-readable range
                    bins = normalize_range(min_, max_ + 1)
                    counts1 = self._lt_ds.counts(c, bins=bins)
                    counts2 = self._rt_ds.counts(c, bins=bins)
                    # Note: index, value of np.histogram has different length
                    bins = bins[:-1]
                self._dt[c]['bins'] = bins
                # stack arrays vertically
                self._dt[c]['counts'] = np.vstack((counts1, counts2))
        return self._dt[column]['bins'], self._dt[column]['counts']

    def describe(self):
        """
        Give descriptive difference between two datasets, which concluded
        relative errors, and jsd divergence.
        Return a panda.DataFrame, whose columns are two dataset's columns, and
        indexes are a array of metrics, e.g. ['err', 'jsd'].
        """
        df1 = self.err()
        df2 = self.jsd()
        return pd.concat([df1, df2])

    def classify(self, label: str, test: pd.DataFrame = None):
        """
        Train two svm classifiers based on data sets, and predict class labels
        for test data. Return both error rates.

        Parameters
        ----------
        label : str
            classifier feature, key is one column in left data frame.
            It supports two-class and multi-class.

        test : {pandas.DataFrame}
            test frame, is test data for machine learning algorithms. If it is
            not provided, it will split 20% of left data frame as test data.

        Returns
        -------
        a DataFrame, e.g.
                         target                         source     target
                      male female                    male female male female
        source male   1    3        or actual male   1    3      1    2
               female 2    4                  female 2    4      3    4
        """
        if (not self._lt_ds[label].categorical or
                not self._rt_ds[label].categorical):
            raise ValueError(
                f'Classifier can\'t run on non-categorical column: {label}')
        from sklearn.metrics import confusion_matrix

        def split_feature_label(df: pd.DataFrame):
            # TODO need improve sub_cols
            sub_cols = [attr for attr in df.columns if attr.startswith(label)]
            if len(sub_cols) == 0:
                return df, None
            is_one_class = len(sub_cols) == 2
            if is_one_class:
                # For one class, there are two sorted values.
                # e.g. ['Yes', 'No'] => [[0, 1],
                #                        [1, 0]]
                # Here it should choose second column to represent this attribute.
                label_ = sub_cols[1]
                return df.drop(sub_cols, axis=1), df[label_]
            else:
                try:
                    # merge multiple column into one: [Name_A, Name_B, ..] => Name
                    _y = df[sub_cols].apply(lambda x: Index(x).get_loc(1),
                                            axis=1)
                    return df.drop(sub_cols, axis=1), _y
                except KeyError as e:
                    print(e)
                    print(sub_cols)
                    print(df[sub_cols])

        # If test dataset is not provided, then split 20% of source dataset for testing.
        if test is None:
            lt, test = train_test_split(self._lt_df, test_size=0.2)
            rt, _ = train_test_split(self._rt_df, test_size=0.2)
        else:
            lt = self._lt_df
            rt = self._rt_df
        lt_x, lt_y = split_feature_label(self._lt_ds.transform(data=lt))
        test_x, test_y = split_feature_label(self._lt_ds.transform(data=test))
        rt_x, rt_y = split_feature_label(self._lt_ds.transform(data=rt))

        # construct svm classifier, and predict
        lt_yp = predict(lt_x, lt_y, test_x)
        rt_yp = predict(rt_x, rt_y, test_x)

        columns = self._lt_ds[label].bins
        labels = range(len(columns))
        # If test dataset has class label, return two expected score.
        if label in test:
            lcm = confusion_matrix(test_y, lt_yp, labels=labels)
            rcm = confusion_matrix(test_y, rt_yp, labels=labels)
            # lcm = lcm.astype('float') / lcm.sum(axis=1)
            # rcm = rcm.astype('float') / rcm.sum(axis=1)
            return (pd.DataFrame(lcm, columns=columns, index=columns),
                    pd.DataFrame(rcm, columns=columns, index=columns))
        # If test dataset don't have class label, return their expected values.
        else:
            cm = confusion_matrix(lt_yp, rt_yp, labels=labels)
            return pd.DataFrame(cm, columns=columns, index=columns)

    def to_html(self):
        pass

    def to_string(self):
        pass


# -----------------------------------------------------------------------------
"""

Base (abstract) class for end-user to implement new machine learning algorithm,
and integrate into this evaluation tool as a plugin.

"""


# ----------------------------------------------------------------------------
"""
Generation of a HTML Report about Anonymization Evaluation Result
"""


class EvaluationReport(object):
    """
    class for Anonymization Evaluation Report.
    It includes:
    basic information, attribute distribution, pair-wise correlation, and
    custom topic, ...
    """
    def __init__(self, title='Evaluation Report'):
        self._title = title
        self._topics = []
        self._content = {}

    def add_part(self, topic, data, description=None):
        """
        Add a new part to evaluation report.

        Parameters
        ----------
        topic : str
            topic of a new part, existing topic includes basic, dist, corr

        data : dict
            data of a new part

        description : str
            detail description of a part
        """
        if topic not in self._topics:
            self._topics.append(topic)
            self._content[topic] = []
        if 'description' not in data and description is not None:
            data['description'] = description
        self._content[topic].append(data)

    def to_html(self, path):
        from mako.template import Template
        import os
        old_cwd = os.getcwd()
        os.chdir(os.path.dirname(__file__))
        template = Template(filename='template/report.html')
        os.chdir(old_cwd)

        def compress_html_content(html: str):
            import re
            value = re.sub(r' //(.*?)\n', '', html)
            value = re.sub(r'\n', ' ', value)
            value = re.sub(r' {2,}', ' ', value)
            return value

        svms = self._content['svm'] if 'svm' in self._content else []
        with open(path, 'w+') as file:
            file.write(compress_html_content(
                template.render(title=self._title,
                                basics=self._content['basic'],
                                dists=self._content['dist'],
                                corrs=self._content['corr'],
                                svms=svms)
            ))
