"""
Metrics for data anonymization
"""
import pandas as pd
import numpy as np
from sklearn.metrics import normalized_mutual_info_score, confusion_matrix
import math


def pairwise_mutual_information(frame: pd.DataFrame):
    """
    Calculate mutual information of a data frame for its all pairwise attributes.

    Returns
    -------
    DataFrame, symmetrical matrix
    """
    columns = frame.columns.sort_values().to_list()
    mi_frame = pd.DataFrame(columns=columns, index=columns, dtype=float)
    for row in mi_frame.columns:
        for col in mi_frame.columns:
            if pd.isna(mi_frame.loc[col, row]):
                mi_frame.loc[row, col] = normalized_mutual_info_score(
                    frame[row].astype(str), frame[col].astype(str),
                    average_method='arithmetic')
            else:
                mi_frame.loc[row, col] = mi_frame.loc[col, row]
    return mi_frame.round(3)


def jensen_shannon_divergence(p, q, base=2):
    """
    Compute the Jensen-Shannon divergence between two 1-D arrays.

    Parameters
    ---------
    p : array
        left probability array
    q : array
        right probability array
    base : numeric, default 2
        logarithm base

    Returns
    -------
    jsd : float
        divergence of p and q
    """
    # If the sum of probability array p or q does not equal to 1, then normalize
    p = np.asarray(p)
    q = np.asarray(q)
    p = p / np.sum(p, axis=0)
    q = q / np.sum(q, axis=0)
    m = (p + q) / 2
    m = sum(- i * math.log(i, base) for i in m if i != 0)
    p = sum(- i * math.log(i, base) for i in p if i != 0)
    q = sum(- i * math.log(i, base) for i in q if i != 0)
    # Follow the equation defined in https://scipy.github.io/devdocs/generated/scipy.spatial.distance.jensenshannon.html
    return round(math.sqrt(m - 0.5 * (p + q)), 4)


def error_rate(y_true, y_pred=None):
    """
    Computer error (mis-classification) rate of one classifier result

    If there is only one parameter, it must be the confusion matrix;
    If there are two parameters, they must be true and predict labels;
    """
    if y_pred is None:
        if isinstance(y_true, pd.DataFrame):
            cm = y_true.values
        else:
            cm = y_true
    else:
        cm = confusion_matrix(y_true, y_pred)
    trace = np.trace(cm)
    sum = np.sum(cm)
    # print(classification_report(test, predict))
    return round((sum - trace) / sum, 4)


def relative_error(x, y):
    """
    Compute relative error of two variables: |x-y|/max(|x|, |y|)
    """
    m = np.maximum(x, y)
    return round(np.average(np.abs(x - y) / np.where(m != 0, m, 1)), 4)
